﻿/*
 * Copyright 2010 Arne Vajhøj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

using System;
using System.Collections.Generic;
using System.IO;

using NUnit.Framework;

namespace Vajhoej.Record.Test
{
    [TestFixture]
    public class TestUtil {
        private const int N = 10000;
        [Test]
        public void TestReadList()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            try
            {
                IList<Data> lst = Util.ReadAll<Data>(typeof(Data), bm);
                Assert.AreEqual(N, lst.Count, "lst size");
                for(int i = 0; i < lst.Count; i++)
                {
                    Assert.AreEqual(258, lst[i].Iv, "iv #" + i);
                    Assert.AreEqual(123.456, lst[i].Xv, 0.0005, "xv #" + i);
                    Assert.AreEqual("ABC     ", lst[i].Sv, "sv #" + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteList()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                IList<Data> lst = new List<Data>();
                for(int i = 0; i < N; i++)
                {
                    Data o = new Data();
                    o.Iv = 258;
                    o.Xv = 123.456;
                    o.Sv = "ABC     ";
                    lst.Add(o);
                }
                byte[] res = Util.WriteAll(typeof(Data), lst);
                Assert.AreEqual(N*b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i % b.Length], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadStream()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            MemoryStream mstm = new MemoryStream(bm);
            try
            {
                IList<Data> lst = new List<Data>();
                Util.ReadAll(typeof(Data), mstm, lst);
                Assert.AreEqual(N, lst.Count, "lst size");
                for(int i = 0; i < lst.Count; i++)
                {
                    Assert.AreEqual(258, lst[i].Iv, "iv #" + i);
                    Assert.AreEqual(123.456, lst[i].Xv, 0.0005, "xv #" + i);
                    Assert.AreEqual("ABC     ", lst[i].Sv, "sv #" + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteStream()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                IList<Data> lst = new List<Data>();
                for(int i = 0; i < N; i++)
                {
                    Data o = new Data();
                    o.Iv = 258;
                    o.Xv = 123.456;
                    o.Sv = "ABC     ";
                    lst.Add(o);
                }
                MemoryStream mstm = new MemoryStream();
                Util.WriteAll(typeof(Data), lst, mstm);
                byte[] res = mstm.ToArray();
                Assert.AreEqual(N*b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i % b.Length], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadHandler()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            MemoryStream mstm = new MemoryStream(bm);
            try
            {
                Util.ObjectHandlerProcess<Data> ohp = delegate(Data o)
                                                      {
                                                          Assert.AreEqual(258, o.Iv, "iv");
                                                          Assert.AreEqual(123.456, o.Xv, 0.0005, "xv");
                                                          Assert.AreEqual("ABC     ", o.Sv, "sv");
                                                      };
                Util.ReadAll<Data>(typeof(Data), mstm, ohp);
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestReadHandlerLambda()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            MemoryStream mstm = new MemoryStream(bm);
            try
            {
                Util.ReadAll<Data>(typeof(Data), mstm, o => { Assert.AreEqual(258, o.Iv, "iv"); Assert.AreEqual(123.456, o.Xv, 0.0005, "xv"); Assert.AreEqual("ABC     ", o.Sv, "sv"); });
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestCopyArrays()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] b2 = { 0x03, 0x01, 0x00, 0x00,
                          0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                          0x41, 0x42, 0x43, 0x58, 0x58, 0x58, 0x58, 0x58 };
            try
            {
                Util.TransformerConvert<Data,Data> tc = delegate(Data o)
                                                        { 
                                                            Data o2 = new Data();
                                                            o2.Iv = o.Iv + 1;
                                                            o2.Xv = o.Xv;
                                                            o2.Sv = o.Sv.Replace(' ', 'X');
                                                            return o2;
                                                        };
                byte[] res = Util.CopyAll<Data,Data>(typeof(Data), b, typeof(Data), tc);
                Assert.AreEqual(b2.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b2[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestCopyArraysLambda()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] b2 = { 0x03, 0x01, 0x00, 0x00,
                          0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                          0x41, 0x42, 0x43, 0x58, 0x58, 0x58, 0x58, 0x58 };
            try
            {
                byte[] res = Util.CopyAll<Data,Data>(typeof(Data), b, typeof(Data), o => new Data { Iv = o.Iv + 1, Xv = o.Xv, Sv = o.Sv.Replace(' ', 'X') });
                Assert.AreEqual(b2.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b2[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestCopyStreams()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] b2 = { 0x03, 0x01, 0x00, 0x00,
                          0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                          0x41, 0x42, 0x43, 0x58, 0x58, 0x58, 0x58, 0x58 };
            try
            {
                MemoryStream inmstm = new MemoryStream(b);
                MemoryStream outmstm = new MemoryStream();
                Util.TransformerConvert<Data,Data> tc = delegate(Data o)
                                                        { 
                                                            Data o2 = new Data();
                                                            o2.Iv = o.Iv + 1;
                                                            o2.Xv = o.Xv;
                                                            o2.Sv = o.Sv.Replace(' ', 'X');
                                                            return o2;
                                                        };
                Util.CopyAll<Data,Data>(typeof(Data),inmstm, typeof(Data), outmstm, tc);
                byte[] res = outmstm.ToArray();
                Assert.AreEqual(b2.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b2[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestCopyStreamsLambda()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] b2 = { 0x03, 0x01, 0x00, 0x00,
                          0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                          0x41, 0x42, 0x43, 0x58, 0x58, 0x58, 0x58, 0x58 };
            try
            {
                MemoryStream inmstm = new MemoryStream(b);
                MemoryStream outmstm = new MemoryStream();
                Util.CopyAll<Data,Data>(typeof(Data),inmstm, typeof(Data), outmstm, o => new Data { Iv = o.Iv + 1, Xv = o.Xv, Sv = o.Sv.Replace(' ', 'X') });
                byte[] res = outmstm.ToArray();
                Assert.AreEqual(b2.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b2[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
    }
    [TestFixture]
    public class TestUtil2
    {
        private const int N = 10000;
        [Test]
        public void TestReadList()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            try
            {
                IList<Data> lst = Util2.ReadAll<Data>(typeof(Data), bm, null, null, null);
                Assert.AreEqual(N, lst.Count, "lst size");
                for(int i = 0; i < lst.Count; i++)
                {
                    Assert.AreEqual(258, lst[i].Iv, "iv #" + i);
                    Assert.AreEqual(123.456, lst[i].Xv, 0.0005, "xv #" + i);
                    Assert.AreEqual("ABC     ", lst[i].Sv, "sv #" + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteList()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                IList<Data> lst = new List<Data>();
                for(int i = 0; i < N; i++)
                {
                    Data o = new Data();
                    o.Iv = 258;
                    o.Xv = 123.456;
                    o.Sv = "ABC     ";
                    lst.Add(o);
                }
                byte[] res = Util2.WriteAll(typeof(Data), lst, null, null, null);
                Assert.AreEqual(N*b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i % b.Length], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadStream()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            MemoryStream mstm = new MemoryStream(bm);
            try
            {
                IList<Data> lst = new List<Data>();
                Util2.ReadAll(typeof(Data), mstm, lst, null, null, null);
                Assert.AreEqual(N, lst.Count, "lst size");
                for(int i = 0; i < lst.Count; i++) {
                    Assert.AreEqual(258, lst[i].Iv, "iv #" + i);
                    Assert.AreEqual(123.456, lst[i].Xv, 0.0005, "xv #" + i);
                    Assert.AreEqual("ABC     ", lst[i].Sv, "sv #" + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestWriteStream()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            try
            {
                IList<Data> lst = new List<Data>();
                for(int i = 0; i < N; i++)
                {
                    Data o = new Data();
                    o.Iv = 258;
                    o.Xv = 123.456;
                    o.Sv = "ABC     ";
                    lst.Add(o);
                }
                MemoryStream mstm = new MemoryStream();
                Util2.WriteAll(typeof(Data), lst, mstm, null, null, null);
                byte[] res = mstm.ToArray();
                Assert.AreEqual(N*b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i % b.Length], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
        [Test]
        public void TestReadHandler()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            MemoryStream mstm = new MemoryStream(bm);
            try
            {
                Util2.ObjectHandlerProcess<Data> ohp = delegate(Data o)
                                                       {
                                                           Assert.AreEqual(258, o.Iv, "iv");
                                                           Assert.AreEqual(123.456, o.Xv, 0.0005, "xv");
                                                           Assert.AreEqual("ABC     ", o.Sv, "sv");
                                                       };
                Util2.ReadAll<Data>(typeof(Data), mstm, ohp, null, null, null);
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestReadHandlerLambda()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            MemoryStream mstm = new MemoryStream(bm);
            try
            {
                Util2.ReadAll<Data>(typeof(Data), mstm, o => { Assert.AreEqual(258, o.Iv, "iv"); Assert.AreEqual(123.456, o.Xv, 0.0005, "xv"); Assert.AreEqual("ABC     ", o.Sv, "sv"); }, null, null, null);
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestCopyArrays()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] b2 = { 0x03, 0x01, 0x00, 0x00,
                          0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                          0x41, 0x42, 0x43, 0x58, 0x58, 0x58, 0x58, 0x58 };
            try
            {
                Util2.TransformerConvert<Data,Data> tc = delegate(Data o)
                                                         {
                                                             Data dres = new Data();
                                                             dres.Iv = o.Iv + 1;
                                                             dres.Xv = o.Xv;
                                                             dres.Sv = o.Sv.Replace(' ', 'X');
                                                             return dres;
                                                         };
                byte[] res = Util2.CopyAll<Data,Data>(typeof(Data), b, typeof(Data), tc, null, null, null);
                Assert.AreEqual(b2.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b2[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestCopyArraysLambda()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] b2 = { 0x03, 0x01, 0x00, 0x00,
                          0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                          0x41, 0x42, 0x43, 0x58, 0x58, 0x58, 0x58, 0x58 };
            try
            {
                byte[] res = Util2.CopyAll<Data,Data>(typeof(Data), b, typeof(Data), o => new Data { Iv = o.Iv + 1, Xv = o.Xv, Sv = o.Sv.Replace(' ', 'X') }, null, null, null);
                Assert.AreEqual(b2.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b2[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestCopyStreams()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] b2 = { 0x03, 0x01, 0x00, 0x00,
                          0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                          0x41, 0x42, 0x43, 0x58, 0x58, 0x58, 0x58, 0x58 };
            try
            {
                MemoryStream inmstm = new MemoryStream(b);
                MemoryStream outmstm = new MemoryStream();
                Util2.TransformerConvert<Data,Data> tc = delegate(Data o)
                                                         {
                                                             Data dres = new Data();
                                                             dres.Iv = o.Iv + 1;
                                                             dres.Xv = o.Xv;
                                                             dres.Sv = o.Sv.Replace(' ', 'X');
                                                             return dres;
                                                         };
                Util2.CopyAll<Data,Data>(typeof(Data), inmstm, typeof(Data), outmstm, tc, null, null, null);
                byte[] res = outmstm.ToArray();
                Assert.AreEqual(b2.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b2[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestCopyStreamsLambda()
        {
            byte[] b = { 0x02, 0x01, 0x00, 0x00,
                         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                         0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
            byte[] b2 = { 0x03, 0x01, 0x00, 0x00,
                          0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                          0x41, 0x42, 0x43, 0x58, 0x58, 0x58, 0x58, 0x58 };
            try
            {
                MemoryStream inmstm = new MemoryStream(b);
                MemoryStream outmstm = new MemoryStream();
                Util2.CopyAll<Data,Data>(typeof(Data), inmstm, typeof(Data), outmstm, o => new Data { Iv = o.Iv + 1, Xv = o.Xv, Sv = o.Sv.Replace(' ', 'X') }, null, null, null);
                byte[] res = outmstm.ToArray();
                Assert.AreEqual(b2.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b2[i], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestVarReadList()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            try
            {
                LengthProvider lenpvd = delegate(object ctx, int n) { return n+1; };
                MaxLengthProvider maxlenpvd = delegate() { return 20; };
                ElementsProvider elmpvd = delegate(object ctx, int n) { return 1; };
                IList<VariableData> lst = Util2.ReadAll<VariableData>(typeof(VariableData), bm, lenpvd, maxlenpvd, elmpvd);
                Assert.AreEqual(N, lst.Count, "lst size");
                for(int i = 0; i < lst.Count; i++)
                {
                    Assert.AreEqual("A", lst[i].S1, "s1");
                    Assert.AreEqual("AB", lst[i].S2, "s2");
                    Assert.AreEqual("ABC",lst[i].S3, "s3");
                    Assert.AreEqual("ABCD", lst[i].S4, "s4");
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestVarReadListLambda()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            try
            {
                IList<VariableData> lst = Util2.ReadAll<VariableData>(typeof(VariableData), bm, (ctx,n) => n+1, () => 20, (ctx,n) => 1);
                Assert.AreEqual(N, lst.Count, "lst size");
                for(int i = 0; i < lst.Count; i++)
                {
                    Assert.AreEqual("A", lst[i].S1, "s1");
                    Assert.AreEqual("AB", lst[i].S2, "s2");
                    Assert.AreEqual("ABC",lst[i].S3, "s3");
                    Assert.AreEqual("ABCD", lst[i].S4, "s4");
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestVarWriteList()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            try
            {
                IList<VariableData> lst = new List<VariableData>();
                for(int i = 0; i < N; i++)
                {
                    VariableData o = new VariableData();
                    o.S1 = "A";
                    o.S2 = "AB";
                    o.S3 = "ABC";
                    o.S4 = "ABCD";
                    lst.Add(o);
                }
                LengthProvider lenpvd = delegate(object ctx, int n) { return n+1; };
                MaxLengthProvider maxlenpvd = delegate() { return 20; };
                ElementsProvider elmpvd = delegate(object ctx, int n) { return 1; };
                byte[] res = Util2.WriteAll(typeof(VariableData), lst, lenpvd, maxlenpvd, elmpvd);
                Assert.AreEqual(N*b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i % b.Length], res[i], "byte " + i);
                }
            }
            catch (RecordException e) 
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestVarWriteListLambda()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            try
            {
                IList<VariableData> lst = new List<VariableData>();
                for(int i = 0; i < N; i++)
                {
                    VariableData o = new VariableData();
                    o.S1 = "A";
                    o.S2 = "AB";
                    o.S3 = "ABC";
                    o.S4 = "ABCD";
                    lst.Add(o);
                }
                byte[] res = Util2.WriteAll(typeof(VariableData), lst, (ctx,n) => n+1, () => 20, (ctx,n) => 1);
                Assert.AreEqual(N*b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i % b.Length], res[i], "byte " + i);
                }
            }
            catch (RecordException e) 
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestVarReadStream()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            MemoryStream mstm = new MemoryStream(bm);
            try
            {
                IList<VariableData> lst = new List<VariableData>();
                LengthProvider lenpvd = delegate(object ctx, int n) { return n+1; };
                MaxLengthProvider maxlenpvd = delegate() { return 20; };
                ElementsProvider elmpvd = delegate(object ctx, int n) { return 1; };
                Util2.ReadAll(typeof(VariableData), mstm, lst, lenpvd, maxlenpvd, elmpvd);
                Assert.AreEqual(N, lst.Count, "lst size");
                for(int i = 0; i < lst.Count; i++)
                {
                    Assert.AreEqual("A", lst[i].S1, "s1");
                    Assert.AreEqual("AB", lst[i].S2, "s2");
                    Assert.AreEqual("ABC",lst[i].S3, "s3");
                    Assert.AreEqual("ABCD", lst[i].S4, "s4");
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestVarReadStreamLambda()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            MemoryStream mstm = new MemoryStream(bm);
            try
            {
                IList<VariableData> lst = new List<VariableData>();
                Util2.ReadAll(typeof(VariableData), mstm, lst, (ctx,n) => n+1, () => 20, (ctx,n) => 1);
                Assert.AreEqual(N, lst.Count, "lst size");
                for(int i = 0; i < lst.Count; i++)
                {
                    Assert.AreEqual("A", lst[i].S1, "s1");
                    Assert.AreEqual("AB", lst[i].S2, "s2");
                    Assert.AreEqual("ABC",lst[i].S3, "s3");
                    Assert.AreEqual("ABCD", lst[i].S4, "s4");
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestVarWriteStream()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            try
            {
                IList<VariableData> lst = new List<VariableData>();
                for(int i = 0; i < N; i++) {
                    VariableData o = new VariableData();
                    o.S1 = "A";
                    o.S2 = "AB";
                    o.S3 = "ABC";
                    o.S4 = "ABCD";
                    lst.Add(o);
                }
                MemoryStream mstm = new MemoryStream();
                LengthProvider lenpvd = delegate(object ctx, int n) { return n+1; };
                MaxLengthProvider maxlenpvd = delegate() { return 20; };
                ElementsProvider elmpvd = delegate(object ctx, int n) { return 1; };
                Util2.WriteAll(typeof(VariableData), lst, mstm, lenpvd, maxlenpvd, elmpvd);
                byte[] res = mstm.ToArray();
                Assert.AreEqual(N*b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i % b.Length], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestVarWriteStreamLambda()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            try
            {
                IList<VariableData> lst = new List<VariableData>();
                for(int i = 0; i < N; i++) {
                    VariableData o = new VariableData();
                    o.S1 = "A";
                    o.S2 = "AB";
                    o.S3 = "ABC";
                    o.S4 = "ABCD";
                    lst.Add(o);
                }
                MemoryStream mstm = new MemoryStream();
                Util2.WriteAll(typeof(VariableData), lst, mstm, (ctx,n) => n+1, () => 20, (ctx,n) => 1);
                byte[] res = mstm.ToArray();
                Assert.AreEqual(N*b.Length, res.Length, "length");
                for(int i = 0; i < res.Length; i++)
                {
                    Assert.AreEqual(b[i % b.Length], res[i], "byte " + i);
                }
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
        [Test]
        public void TestVarReadHandler()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            MemoryStream mstm = new MemoryStream(bm);
            try
            {
                Util2.ObjectHandlerProcess<VariableData> ohp = delegate(VariableData o)
                                                               {
                                                                   Assert.AreEqual("A", o.S1, "s1");
                                                                   Assert.AreEqual("AB", o.S2, "s2");
                                                                   Assert.AreEqual("ABC", o.S3, "s3");
                                                                   Assert.AreEqual("ABCD", o.S4, "s4");
                                                               };
                LengthProvider lenpvd = delegate(object ctx, int n) { return n+1; };
                MaxLengthProvider maxlenpvd = delegate() { return 20; };
                ElementsProvider elmpvd = delegate(object ctx, int n) { return 1; };
                Util2.ReadAll<VariableData>(typeof(VariableData), mstm, ohp, lenpvd, maxlenpvd, elmpvd);
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#if DN35
        [Test]
        public void TestVarReadHandlerLambda()
        {
            byte[] b = { 0x41,
                         0x41, 0x42,
                         0x41, 0x42, 0x43,
                         0x41, 0x42, 0x43, 0x44 };
            byte[] bm = new byte[N*b.Length];
            for(int i = 0; i < N; i++)
            {
                Array.Copy(b, 0, bm, i*b.Length, b.Length);
            }
            MemoryStream mstm = new MemoryStream(bm);
            try
            {
                Util2.ReadAll<VariableData>(typeof(VariableData), mstm, o => { Assert.AreEqual("A", o.S1, "s1"); Assert.AreEqual("AB", o.S2, "s2"); Assert.AreEqual("ABC", o.S3, "s3"); Assert.AreEqual("ABCD", o.S4, "s4"); }, (ctx,n) => n+1, () => 20, (ctx,n) => 1);
            }
            catch (RecordException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
            catch (IOException e)
            {
                Assert.Fail("Unexpected exception: " + e);
            }
        }
#endif
    }
}